/* compile and link with gcc USBAudioRecorder.c -o RunImage -lasound -lmath */

#include <stdio.h>
#include <alsa/asoundlib.h>
#include <termios.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <math.h>

#define ALSA_PCM_NEW_HW_PARAMS_API

static double dbfs [129] = { -60.0, -42.1, -36.1, -32.5, -30.0, -28.1, -26.5, -25.2, -24.0, -23.0, -22.1, -21.2, -20.5, -19.8, -19.2, -18.6, -18.0, -17.5, -17.0, -16.5, -16.1, -15.6, -15.2, -14.8, -14.5, -14.1, -13.8, -13.4, -13.1, -12.8, -12.5, -12.2, -12.0, -11.7, -11.4, -11.2, -11.0, -10.7, -10.5, -10.3, -10.0,  -9.8,  -9.6,  -9.4,  -9.2,  -9.0,  -8.8,  -8.6,  -8.5,  -8.3,  -8.1,  -7.9,  -7.8,  -7.6,  -7.4,  -7.3,  -7.1,  -7.0,  -6.8,  -6.7,  -6.5,  -6.4,  -6.2,  -6.1,  -6.0,  -5.8,  -5.7,  -5.6,  -5.4,  -5.3,  -5.2,  -5.1,  -4.9,  -4.8,  -4.7,  -4.6,  -4.5,  -4.3,  -4.2,  -4.1,  -4.0,  -3.9,  -3.8,  -3.7,  -3.6,  -3.5,  -3.4,  -3.3,  -3.2,  -3.1,  -3.0,  -2.9,  -2.8,  -2.7,  -2.6,  -2.5,  -2.4,  -2.3,  -2.3,  -2.2,  -2.1,  -2.0,  -1.9,  -1.8,  -1.7,  -1.7,  -1.6,  -1.5,  -1.4,  -1.3,  -1.2,  -1.2,  -1.1,  -1.0,  -0.9,  -0.9,  -0.8,  -0.7,  -0.6,  -0.6,  -0.5,  -0.4,  -0.3,  -0.3,  -0.2,  -0.1,  -0.1,   0.0};



snd_pcm_t *alsa_handle;
snd_pcm_hw_params_t *alsa_params;
snd_pcm_uframes_t alsa_frames;
unsigned char *alsa_buffer;
unsigned char *outbuffer;
char AlsaDevString[128];
int alsa_rc, alsa_size, alsa_dir;
unsigned int alsa_val;

int grab_count, sec_count, half_sec;

long payload, max_payload;

long slat;

double db0;

long SampleRate,ByteRate,BlockAlign;
long BitsPerSample,BytesPerSample;
long BytesPerTransfer;
long ls2max, rs2max;

long InSampleRate,InBlockAlign,InByteRate;

FILE* outfile;

char outfilename[1024];
char rstate[8];

double ls_max, rs_max, rs_db, ls_db, ls_ppm, rs_ppm;
int secs, mins, recording, quit, fixpayload,converting,bi,bo;
char keygot;

char appdirname[1024];
char settingsfilename[1024];
char outfilename[1024];
char outdirname[1024];

void reportline(void);


void write4int(long);
void write2int(long);
void write4(char*);
void write_output_header(long);

void setup_sound(void);
void grab_sound(void);
void shutdown_sound(void);
void process_grab(void);
void get_peaks(void);
void lineback(int);
int kbhit(void);
void changemode(int);
void fixheader(void);
void byteconvert(void);

void read_settings(void);

int main(int argc,char*argv[])
{
  
  printf("  ROXAudioRecorder Ver 0.55  26th June 2014\n");
  printf("  =========================================\n");
  
  printf("  Simple program for recording stereo audio.\n\n");
  
  sprintf(appdirname,"%s",argv[0]);
   
  slat=(long)strrchr(appdirname,'/')-(long)appdirname;
  appdirname[slat]=(char)0;
  /* printf("App at %s\n",appdirname); */
  sprintf(settingsfilename,"%s/Settings",appdirname);
  printf("Read settings from %s\n",settingsfilename);
  
  read_settings();
  sprintf(outfilename,"%s/NewRecording.wav",outdirname);
  printf("Output to file = %s\n",outfilename);
  
  max_payload=max_payload*1024*1024;
  BytesPerSample=BitsPerSample/8;
  BlockAlign=2*BytesPerSample;
  ByteRate=SampleRate*2*BytesPerSample;
  bi=BytesPerTransfer;  bo=BytesPerSample;
  InSampleRate=SampleRate;
  if(BytesPerTransfer==BytesPerSample)
  {
    
    InBlockAlign=BlockAlign;
    InByteRate=ByteRate;
    converting=0;
    printf("No sample size conversion %d = %d\n",bi,bo);
  }
  else
  {
    InBlockAlign=2*bi;
    InByteRate=SampleRate*InBlockAlign;
    converting=1;
    printf("Sample size conversion %d -> %d bytes per sample\n",bi,bo);
  }
  
  /* printf("Sample rate = %ld  Bits per sample = %ld\n",SampleRate,BitsPerSample); */
  printf("In:  ByteRate = %ld   BytesPerTransfer = %ld  BlockAlign = %ld\n",InByteRate,BytesPerTransfer,InBlockAlign);
  printf("Out: ByteRate = %ld   BytesPerSample   = %ld  BlockAlign = %ld\n\n",ByteRate,BytesPerSample,BlockAlign);
  /* printf("Max payload set to %ld Bytes\n\n\n",max_payload); */
  printf(" Press 'R' to start the recording.   Press 'Q' to quit\n");
  printf(" =====================================================\n\n");
  
  sprintf(rstate,"Ready");
  setup_sound();
  printf("Sound setup\n");
  
  grab_count=0; sec_count=0; secs=0; mins=0; half_sec=0; payload=0; keygot=0;
  recording=0; quit=0; fixpayload=0; db0=log10(32768.0);
  ls2max=2; rs2max=2; ls_ppm=-84.0; rs_ppm=-84.0;
  changemode(1);
  do
  {
    grab_sound();
    process_grab();
    reportline();
    if(kbhit()>0)
    {
      keygot=getchar();
      switch(keygot)
      {
        case 'q': case 'Q':
          quit=1;
          printf("\nQuit!\n\n");
          sprintf(rstate,"Stop ");
          if(recording==1)
          {
            fixpayload=1;
          }
          break;
          case 'r': case 'R':
          if (recording==0)
          {
            recording=1;
            printf("\nRecording started!\n");
            sprintf(rstate,"Rec  ");
            outfile=fopen(outfilename,"wt");
            write_output_header(10*ByteRate);
          }
          break;
       }
    }
    else
    {
      keygot=' ';
    }
    
    if(recording==1)
    {
      grab_count++;
      half_sec++;
      payload+=ByteRate/2;
      if(half_sec==2)
      {
        sec_count++;
        half_sec=0;
        secs++;
        if(secs==60)
        {
          secs=0; mins++;
        }
      }
      
      if(converting==0)
      {
        fwrite(alsa_buffer,sizeof(char),ByteRate/2,outfile);
      }
      else
      {
        byteconvert();
        fwrite(outbuffer,sizeof(char),ByteRate/2,outfile);
      }
    }
    
    if(payload>=max_payload)
    {
      quit=1; fixpayload=1;
      printf("Max payload reached!\n");
    }
    
  } while (quit==0);
  
  reportline();
  changemode(0);
  shutdown_sound();
  printf("\nSound shutdown\n");
  if(recording==1)
  {
    fclose(outfile);
    printf("File closed\n");
    printf("Output saved to %s\n",outfilename);
    if(fixpayload==1)
    {
      printf("fixing playload to %ld\n",payload);
      fixheader();
    }
    
  }
  printf(" *** Done! ***\n");
}

void reportline(void)
{
  printf("%s %3dm %02d.%ds => %6.1lf MB  L/R = %5.1lf  %5.1lf   ppm  %5.1lf  %5.1lf  dB\r",rstate,mins,secs,5*half_sec,9.53674e-7*(double)payload,ls_db,rs_db,ls_ppm,rs_ppm);
}

void read_settings(void)
{
  FILE* setfile;
  setfile=fopen(settingsfilename,"rt");
  fscanf(setfile,"%s",outdirname);
  printf("Output to dir = %s\n",outdirname);
  fscanf(setfile,"%s",AlsaDevString);
  printf("Using Alsa device = %s\n",AlsaDevString);
  fscanf(setfile,"%ld %ld %ld",&SampleRate,&BitsPerSample,&BytesPerTransfer);
  printf("Sample Rate = %ld   Bits per sample = %ld\n",SampleRate,BitsPerSample);
  fscanf(setfile,"%ld",&max_payload);
  printf("Maximum payload cutoff = %ld MB\n",max_payload);
  fclose(setfile);
}

void fixheader(void)
{
  unsigned long pxbuf[1];
  outfile=fopen(outfilename,"r+");
  fseek(outfile,4,SEEK_SET);
  
  pxbuf[0]=payload+36;
  fwrite(pxbuf,sizeof(long),(long)1,outfile);
  fseek(outfile,(long)40,SEEK_SET);
  pxbuf[0]=payload;
  fwrite(pxbuf,sizeof(long),(long)1,outfile);
  fclose(outfile);
}

void lineback(int nback)
{
  int backcount=0;
  do
  {
    printf("\b");
    backcount++;
  } while (backcount<nback);
  printf("\r");
}

void get_peaks(void)
{
  char cgot;
  double ls, rs;
  int pk,vk;
  long lsnow, rsnow;
  short* lspoint;
  short* rspoint;
  
  long gpend;
  gpend=InByteRate/2 - 2*InBlockAlign;
  
  
  ls2max=0; rs2max=0;
  
  pk=BytesPerTransfer-1;
  do
  {
    
    
    lspoint=(short*)(pk-1+(long)alsa_buffer);
    lsnow=(long)lspoint[0];
    if(lsnow<0) lsnow=(-lsnow);
    if(lsnow>ls2max) ls2max=lsnow;
    pk+=BytesPerTransfer;
    
    rspoint=(short*)(pk-1+(long)alsa_buffer);
    rsnow=(long)rspoint[0];
    if(rsnow<0) rsnow=(-rsnow);
    if(rsnow>rs2max) rs2max=rsnow;
    pk+=BytesPerTransfer;
    
  } while (pk<gpend);
  
  if (ls2max<2) ls2max=2;
  if (rs2max<2) rs2max=2;
 
  ls_ppm-=1.0; rs_ppm-=1.0;
  
  ls_db=20.0*( log10((double)ls2max)-db0);
  if(ls_db>ls_ppm) ls_ppm=ls_db;
  
  rs_db=20.0*( log10((double)rs2max)-db0);
  if(rs_db>rs_ppm) rs_ppm=rs_db;
}


void byteconvert(void)
{
  int cout,cin,cio,jump;
  long outsize;
  outsize=ByteRate/2; jump=bi-bo;
  cout=0; cin=bi-bo; cio=0;
  
  do
  {
    do
    {
      outbuffer[cout]=alsa_buffer[cin];
      cout++;
      cin++;
      cio++;
    } while (cio<bo);
    cin+=jump;
    cio=0;
  } while (cout<outsize);
  
  if(cout!=outsize)
  {
    printf("convert error %d %ld\n",cout,outsize);
  }
}


void process_grab(void)
{
  printf("\r");
  get_peaks();
  printf("\n");
}

void grab_sound(void)
{
  alsa_rc = snd_pcm_readi(alsa_handle, alsa_buffer, alsa_frames);
    if (alsa_rc == -EPIPE) {
      /* EPIPE means overrun */
      fprintf(stderr, "overrun occurred\n");
      
      snd_pcm_prepare(alsa_handle);
    } else if (alsa_rc < 0) {
      fprintf(stderr,
              "error from read: %s\n", 
              snd_strerror(alsa_rc));
    } else if (alsa_rc != (int)alsa_frames) {
      fprintf(stderr, "short read, read %d frames\n", alsa_rc);
    }
}


void shutdown_sound(void)
{
  snd_pcm_drain(alsa_handle);
  snd_pcm_close(alsa_handle);
  free(alsa_buffer);
}


void setup_sound(void)
{
  
  /* Open PCM device for recording (capture). */
  alsa_rc = snd_pcm_open(&alsa_handle, AlsaDevString,
                    SND_PCM_STREAM_CAPTURE, 0);
  if (alsa_rc < 0) {
    fprintf(stderr, 
            "unable to open pcm device: %s\n",
            snd_strerror(alsa_rc));
    exit(1);
  }
  
  /* Allocate a hardware parameters object. */
  snd_pcm_hw_params_alloca(&alsa_params);

  /* Fill it in with default values. */
  snd_pcm_hw_params_any(alsa_handle, alsa_params);

  /* Set the desired hardware parameters. */

  /* Interleaved mode */
  snd_pcm_hw_params_set_access(alsa_handle, alsa_params,
                      SND_PCM_ACCESS_RW_INTERLEAVED);

  /* Signed little-endian format */
  
  if (BytesPerTransfer==2)
  {
    snd_pcm_hw_params_set_format(alsa_handle, alsa_params,SND_PCM_FORMAT_S16_LE);
    printf("Set s16_le format  n");
  }
  
  if (BytesPerTransfer==3)
  {
    snd_pcm_hw_params_set_format(alsa_handle, alsa_params,SND_PCM_FORMAT_S24_LE);
    printf("Set s24_le format  ");
  }
  
  if (BytesPerTransfer==4)
  {
    snd_pcm_hw_params_set_format(alsa_handle, alsa_params,SND_PCM_FORMAT_S32_LE);
    printf("Set s32_le format  ");
  }
                              

  /* Two channels (stereo) */
  snd_pcm_hw_params_set_channels(alsa_handle, alsa_params, 2);
  printf("2 channels ");

  /* bits/second sampling rate */
  alsa_val = InSampleRate;
  
  /* printf("alsa val sample rate %d\n",alsa_val); */
  
  snd_pcm_hw_params_set_rate_near(alsa_handle, alsa_params, 
                                  &alsa_val, &alsa_dir);

  /* Set period size for half a second of frames. */
  alsa_frames = InSampleRate/2;
  
  /* printf("alsa frame %d\n",(int)alsa_frames); */
  
  snd_pcm_hw_params_set_period_size_near(alsa_handle, 
                              alsa_params, &alsa_frames, &alsa_dir);
  
  
  
  /* Write the parameters to the driver */
  alsa_rc = snd_pcm_hw_params(alsa_handle, alsa_params);
  if (alsa_rc < 0) {
    fprintf(stderr,
            "unable to set hw parameters: %s\n",
            snd_strerror(alsa_rc));
    exit(1);
  }

  /* Use a buffer large enough to hold one period */
  snd_pcm_hw_params_get_period_size(alsa_params,
                                      &alsa_frames, &alsa_dir);
  alsa_size = alsa_frames * InBlockAlign; /* X bytes/sample, 2 channels */
  
  
  alsa_buffer = (char *) malloc(alsa_size);
  
  if(converting==1)
  {
    outbuffer = (char *) malloc(ByteRate/2);
  }
  
  snd_pcm_hw_params_get_period_time(alsa_params,
                                         &alsa_val, &alsa_dir);
  
  
}

void changemode(int dir)
{
   static struct termios oldt, newt;

   if ( dir == 1 )
   {
     tcgetattr( STDIN_FILENO, &oldt);
     newt = oldt;
     newt.c_lflag &= ~( ICANON | ECHO );
     tcsetattr( STDIN_FILENO, TCSANOW, &newt);
   }
   else
     tcsetattr( STDIN_FILENO, TCSANOW, &oldt);
}


int kbhit (void)
{
   struct timeval tv;
   fd_set rdfs;

   tv.tv_sec = 0;
   tv.tv_usec = 0;

   FD_ZERO(&rdfs);
   FD_SET (STDIN_FILENO, &rdfs);

   select(STDIN_FILENO+1, &rdfs, NULL, NULL, &tv);
   return FD_ISSET(STDIN_FILENO, &rdfs);

}

void write4int(long inv)
{
  char cop1,cop2,cop3,cop4;
  
  cop1=(unsigned char)(inv&255);
  cop2=(unsigned char)((inv>>8)&255);
  cop3=(unsigned char)((inv>>16)&255);
  cop4=(unsigned char)((inv>>24)&255);
  
  fprintf(outfile,"%c",cop1);
  fprintf(outfile,"%c",cop2);
  fprintf(outfile,"%c",cop3);
  fprintf(outfile,"%c",cop4);
}


void write2int(long inv)
{
  unsigned char cop1,cop2;
  cop1=(unsigned char)(inv&255);
  cop2=(unsigned char)((inv>>8)&255);
  fprintf(outfile,"%c",cop1);
  fprintf(outfile,"%c",cop2);
}

void write4(char* outstring)
{
  int isp;
  
  isp=0;
  do
  {
    fprintf(outfile,"%c",outstring[isp]);
    isp++;
  } while (isp<4);
  
}


void write_output_header(long plo)
{
  
  
  write4("RIFF\0");
  write4int(plo+36);
  write4("WAVE\0");
  write4("fmt \0");
  write4int(16);
  write2int(1);
  write2int(2);
  write4int(SampleRate);
  write4int(ByteRate);
  write2int(BlockAlign);
  write2int(BitsPerSample);
  write4("data\0");
  write4int(plo);
  
}




